import csv
import os
import re
import sys

# Use typing library for type hinting the file type, after importing IO
from typing import IO

#from tabulate import tabulate

from bio import DNA_obj


def main():
    """
    The main function of the script.
    It provides directions for using the script
    and runs the functions that emulate the Central Dogma of Biology for a given DNA sequence.
    It also runs functions that calculate phosphodiester bonds, hydrogen bonds and GC content
    for every DNA sequence's corresponding double stranded DNA molecule.

    :param: None
    :return: None
    """

    print("This is Central Dogma Script.")
    print("This script simulates the Central Dogma of Biology.")
    print("It uses DNA sequences formated as plain text and outputs:")
    print("1. The complementary DNA sequences.")
    print("2. The mRNA sequences.")
    print("3. The polypeptide sequences.")
    print("4 The number of Hydrogen and Phosphodiester bonds.")
    print("The GC content of the corresponding double stranded DNA molecules.")

    while True:
        response = (
            input("Existing output files may be overwritten.\nContinue? (Y/N)")
            .strip()
            .upper()
        )
        if response == "Y":
            break
        elif response == "N":
            sys.exit()
        else:
            continue
    if len(sys.argv) == 2:
        if os.path.isfile(sys.argv[1]):
            print("Your data file exists!")
            print(
                "Read the dogma.txt file for Replication, Transcription and Translation data., plus various calculations for each DNA sequence."
            )
            print("Your translation data is written in the peptides.csv file.")
            print("For tabular format the peptides_tab.txt was created.")
            if os.path.isfile("./dogma.txt"):
                os.remove("./dogma.txt")
            create_csv_dogma_files(sys.argv[1], "peptides.csv", "dogma.txt")
            #create_tab_table("peptides.csv", "peptides_tab.txt")
        else:
            print("Coding Sequence file not found.")
            print("Check that the path and the name of your sequence file are correct,")
            print(
                "or run the program without arguments and input a coding DNA sequence manually."
            )
    else:
        print(
            "If you want to load a data file for multiple coding sequences run this script as:"
        )
        print("python3 central_dogma.py /path/name_of_file.extension")
        if os.path.isfile("./dogma_manual.txt"):
            os.remove("./dogma_manual.txt")
        while True:
            d_seq = input("Enter your coding DNA sequence:\n")
            print("Enter the direction of your sequence:")
            while True:
                print("1 for 5->3")
                print("2 for 3->5")
                d = input("Enter your direction: ")
                if d != "1" and d != "2":
                    continue
                else:
                    break
            dna_object = set_coding_sequence(d_seq, d)
            print(dna_object)
            bases_dic = count_bases(dna_object.seq)
            print(f"Hydrogen bonds = {compute_hydro_bonds(bases_dic)}")
            print(f"Phosphodiester bonds = {compute_phospho_bonds(dna_object.seq)}")
            print(f"GC content = {compute_gc_content(bases_dic)}")
            create_dogma_file("dogma_manual.txt", dna_object, "a")
            print("The output was added to the dogma_manual.txt file.")
            response = input("Do you want to continue? (Y/N)").strip().upper()
            if response == "Y":
                continue
            else:
                break


def set_coding_sequence(c_seq: str, direct: str = "1") -> DNA_obj:
    """
    Instantiates a new DNA sequence object.
    It assigns values to all DNA sequence object's parameters, using the DNA_obj class methods and returns the object.

    :param c_seq: A string that holds a candidate DNA sequence
    :type c_seq: str
    :param direct: A string that holds the orientation of the DNA sequence. "1" is for 5'->3' and "2" is for 3'->5' (default is "1").
    :type direct: str
    :return: A DNA sequence object made by the custom class DNA_obj
    :rtype: DNA_obj
    """
    dna_obj = DNA_obj(c_seq, direct)
    dna_obj.set_seq_dir()
    dna_obj.comp()
    dna_obj.transcription()
    dna_obj.translation()
    return dna_obj


def create_dogma_file(out_filename: str, d_object: DNA_obj, write_mode: str):
    """
    Creates a plain text output file, providing different writing modes.
    It is used in `create_csv_dogma_files` function.

    :param out_filename: A string that holds the name of the plain text output file
    :type out_filename: str
    :param d_object: A DNA sequence object made by the custom class DNA_obj
    :type d_object: DNA_obj
    :param write_mode: A string that specifies the writing mode
    :type write_mode: str
    """
    with open(out_filename, write_mode, encoding="utf-8") as out_file:
        out_file.write(str(d_object) + "\n")
        bases_d = count_bases(d_object.seq)
        out_file.write(f"Hydrogen bonds = {compute_hydro_bonds(bases_d)}\n")
        out_file.write(
            f"Phosphodiester bonds = {compute_phospho_bonds(d_object.seq)}\n"
        )
        out_file.write(f"gc content = {compute_gc_content(bases_d)}\n\n")


def create_csv_dogma_files(in_filename: str, out_filename1: str, out_filename2: str):
    # open files with specified encoding and newline="" for platform compatibility
    # when reading the csv file in a rows dictionary,
    # newline="" specifies that if a newline exists,
    # it will not be stored in each dictionary's row.
    """
    Creates the csv and the plain text output text files
    For writing the plain text file, it uses the `create_dogma_file` function

    :param in_filename: A string that holds the name of the input plain text file, that contains the DNA sequences
    :type in_filename: str
    :param out_filename1: A string that holds the name of the csv output file
    :type out_filename1: str
    :param out_filename2: A string that holds the name of the plain text output file
    :type out_filename2: str
    """
    with open(out_filename1, "w", encoding="utf-8", newline="") as out_file:
        writer = csv.writer(out_file)
        writer.writerow(
            ["DNA Coding Sequence 5' -> 3'", "Peptide Sequence H₂N -> COOH"]
        )
        with open(in_filename, "r", encoding="utf-8", newline="") as in_file:
            for line in in_file:
                dna_object = set_coding_sequence(line)
                writer.writerow(
                    [
                        dna_object.justify_nucleotide_seq(dna_object.seq),
                        dna_object.peptide_seq,
                    ]
                )
                create_dogma_file(out_filename2, dna_object, "a")


# Create a list of rows from the peptides.csv output file
# in order to feed the tabulate function in create_tab_table
#def create_table_list(file: IO[str]) -> list[list[str]]:
#    """
#    Creates a list of rows from a csv file
#
#    :param file: A file object that holds the csv input file
#    :type file: file
#    :return: A list of rows
#    :rtype: list
#    """
#    list_of_rows: list[list[str]] = []
#    reader = csv.reader(file)
#    for row in reader:
#        list_of_rows.append(row)
#    return list_of_rows


#def create_tab_table(in_filename: str, out_filename: str):
#    """
#    Creates an output text file, that contains the csv table in tabular format
#
#    :param in_filename: A string that holds the file name of the input csv file
#    :type in_filename: str
#    :param out_filename: A string that holds the file name of the output tabular text file
#    :type out_filename: str
#    """
#    with open(out_filename, "w", encoding="utf-8", newline="") as output_file:
#        with open(in_filename, "r", encoding="utf-8", newline="") as input_file:
#            output_file.write(
#                tabulate(
#                    create_table_list(input_file), headers="firstrow", tablefmt="grid"
#                )
#            )


def count_bases(dna_seq: str):
    """
    Count the number of nitrogenous bases of a DNA sequence

    :param dna_seq: A string that holds the DNA sequence
    :type dna_seq: str
    :raise ValueError: If the DNA sequence contains invalid characters
    :return: A dictionary that holds the counts of nitrogenous bases
    :rtype: dict
    """
    bases_dict = {"A": 0, "T": 0, "G": 0, "C": 0}
    for c in dna_seq:
        if c == "A":
            bases_dict["A"] += 1
        elif c == "T":
            bases_dict["T"] += 1
        elif c == "G":
            bases_dict["G"] += 1
        elif c == "C":
            bases_dict["C"] += 1
        else:
            raise ValueError("Invalid Sequence!")
    return bases_dict


def compute_hydro_bonds(b_dict) -> int:
    """
    Compute the number of hydrogen bonds in the double stranded DNA

    :param b_dict: A dictionary that holds the counts of DNA bases
    :type b_dict: dict
    :return: The number of hydrogen bonds in the double stranded DNA
    :rtype: int
    """
    hydrogen_bonds = 2 * (b_dict["A"] + b_dict["T"]) + 3 * (b_dict["G"] + b_dict["C"])
    return hydrogen_bonds


def compute_phospho_bonds(dna_seq: str) -> int:
    """
    Compute the number of phosphodiester bonds in the double stranded DNA

    :param dna_seq: The DNA sequence
    :type dna_seq: str
    :raise ValueError: If the DNA sequence contains invalid characters
    :return: The number of phosphodiester bonds in the double stranded DNA
    :rtype: int
    """
    if re.search(r"[^atgcATGC]", dna_seq):
        raise ValueError("Invalid Sequence!")
    if 0 <= len(dna_seq) <= 1:
        return 0
    return 2 * len(dna_seq) - 2


def compute_gc_content(b_dict) -> float:
    """
    Compute the gc content of the double stranded DNA

    :param b_dict: A dictionary that holds the counts of DNA bases
    :type b_dict: dict
    :return: The gc content of the double stranded DNA. If the DNA sequence has 0 length gc content will be 0
    :rtype: float
    """
    at = b_dict["A"] + b_dict["T"]
    gc = b_dict["G"] + b_dict["C"]
    if at + gc == 0:
        return 0
    return gc / (gc + at)


if __name__ == "__main__":
    main()
